@@ -1,21 +1,21 @@
 import os
 import os.path as op
 import platform
-import urlparse
 import re
 import shutil
 
 from operator import itemgetter
-
 from werkzeug import secure_filename
 
-from flask import flash, url_for, redirect, abort, request
+from flask import flash, url_for, redirect, abort, request, send_file
 
+from wtforms import fields, validators
+
+from flask.ext.admin import form, helpers
+from flask.ext.admin._compat import urljoin, as_unicode
 from flask.ext.admin.base import BaseView, expose
 from flask.ext.admin.actions import action, ActionsMixin
 from flask.ext.admin.babel import gettext, lazy_gettext
-from flask.ext.admin import form
-from flask.ext import wtf
 
 
 class NameForm(form.BaseForm):
@@ -24,13 +24,13 @@ class NameForm(form.BaseForm):
 
         Validates if provided name is valid for *nix and Windows systems.
     """
-    name = wtf.TextField()
+    name = fields.TextField()
 
     regexp = re.compile(r'^(?!^(PRN|AUX|CLOCK\$|NUL|CON|COM\d|LPT\d|\..*)(\..+)?$)[^\x00-\x1f\\?*:\";|/]+$')
 
     def validate_name(self, field):
         if not self.regexp.match(field.data):
-            raise wtf.ValidationError(gettext('Invalid directory name'))
+            raise validators.ValidationError(gettext('Invalid directory name'))
 
 
 class UploadForm(form.BaseForm):
@@ -38,39 +38,38 @@ class UploadForm(form.BaseForm):
         File upload form. Works with FileAdmin instance to check if it is allowed
         to upload file with given extension.
     """
-    upload = wtf.FileField(lazy_gettext('File to upload'))
+    upload = fields.FileField(lazy_gettext('File to upload'))
 
     def __init__(self, admin):
         self.admin = admin
 
-        super(UploadForm, self).__init__()
+        super(UploadForm, self).__init__(helpers.get_form_data())
 
     def validate_upload(self, field):
-        if not self.upload.has_file():
-            raise wtf.ValidationError(gettext('File required.'))
+        if not self.upload.data:
+            raise validators.ValidationError(gettext('File required.'))
 
         filename = self.upload.data.filename
 
         if not self.admin.is_file_allowed(filename):
-            raise wtf.ValidationError(gettext('Invalid file type.'))
+            raise validators.ValidationError(gettext('Invalid file type.'))
 
 
 class EditForm(form.BaseForm):
-    content = wtf.TextAreaField(lazy_gettext('Content'),
-                                [wtf.validators.required()])
+    content = fields.TextAreaField(lazy_gettext('Content'),
+                                   (validators.required(),))
 
 
 class FileAdmin(BaseView, ActionsMixin):
     """
         Simple file-management interface.
 
-        Requires two parameters:
-
         :param path:
             Path to the directory which will be managed
-        :param url:
-            Base URL for the directory. Will be used to generate
-            static links to the files.
+        :param base_url:
+            Optional base URL for the directory. Will be used to generate
+            static links to the files. If not defined, a route will be created
+            to serve uploaded files.
 
         Sample usage::
 
@@ -86,6 +85,11 @@ class FileAdmin(BaseView, ActionsMixin):
         Is file upload allowed.
     """
 
+    can_download = True
+    """
+        Is file download allowed.
+    """
+
     can_delete = True
     """
         Is file deletion allowed.
@@ -151,7 +155,12 @@ class FileAdmin(BaseView, ActionsMixin):
         Edit template
     """
 
-    def __init__(self, base_path, base_url,
+    upload_form = UploadForm
+    """
+        Upload form class
+    """
+
+    def __init__(self, base_path, base_url=None,
                  name=None, category=None, endpoint=None, url=None,
                  verify_path=True):
         """
@@ -173,7 +182,7 @@ class FileAdmin(BaseView, ActionsMixin):
                 Verify if path exists. If set to `True` and path does not exist
                 will raise an exception.
         """
-        self.base_path = base_path
+        self.base_path = as_unicode(base_path)
         self.base_url = base_url
 
         self.init_actions()
@@ -181,13 +190,13 @@ class FileAdmin(BaseView, ActionsMixin):
         self._on_windows = platform.system() == 'Windows'
 
         # Convert allowed_extensions to set for quick validation
-        if (self.allowed_extensions
-            and not isinstance(self.allowed_extensions, set)):
+        if (self.allowed_extensions and
+            not isinstance(self.allowed_extensions, set)):
             self.allowed_extensions = set(self.allowed_extensions)
 
         # Convert editable_extensions to set for quick validation
-        if (self.editable_extensions
-            and not isinstance(self.editable_extensions, set)):
+        if (self.editable_extensions and
+            not isinstance(self.editable_extensions, set)):
             self.editable_extensions = set(self.editable_extensions)
 
         # Check if path exists
@@ -281,7 +290,7 @@ class FileAdmin(BaseView, ActionsMixin):
         """
         file_data.save(path)
 
-    def _get_dir_url(self, endpoint, path, **kwargs):
+    def _get_dir_url(self, endpoint, path=None, **kwargs):
         """
             Return prettified URL
 
@@ -310,10 +319,10 @@ class FileAdmin(BaseView, ActionsMixin):
                 Static file path
         """
         if self.is_file_editable(path):
-            return url_for(".edit", path=path)
+            route = '.edit'
         else:
-            base_url = self.get_base_url()
-            return urlparse.urljoin(base_url, path)
+            route = '.download'
+        return url_for(route, path=path)
 
     def _normalize_path(self, path):
         """
@@ -406,6 +415,17 @@ class FileAdmin(BaseView, ActionsMixin):
         """
         pass
 
+    def _save_form_files(self, directory, path, form):
+        filename = op.join(directory,
+                           secure_filename(form.upload.data.filename))
+
+        if op.exists(filename):
+            flash(gettext('File "%(name)s" already exists.', name=filename),
+                  'error')
+        else:
+            self.save_file(filename, form.upload.data)
+            self.on_file_upload(directory, path, filename)
+
     @expose('/')
     @expose('/b/<path:path>')
     def index(self, path=None):
@@ -418,6 +438,10 @@ class FileAdmin(BaseView, ActionsMixin):
         # Get path and verify if it is valid
         base_path, directory, path = self._normalize_path(path)
 
+        if not self.is_accessible_path(path):
+            flash(gettext('Permission denied.'))
+            return redirect(self._get_dir_url('.index'))
+
         # Get directory listing
         items = []
 
@@ -431,8 +455,10 @@ class FileAdmin(BaseView, ActionsMixin):
 
         for f in os.listdir(directory):
             fp = op.join(directory, f)
+            rel_path = op.join(path, f)
 
-            items.append((f, op.join(path, f), op.isdir(fp), op.getsize(fp)))
+            if self.is_accessible_path(rel_path):
+                items.append((f, rel_path, op.isdir(fp), op.getsize(fp)))
 
         # Sort by name
         items.sort(key=itemgetter(0))
@@ -475,24 +501,41 @@ class FileAdmin(BaseView, ActionsMixin):
             flash(gettext('File uploading is disabled.'), 'error')
             return redirect(self._get_dir_url('.index', path))
 
-        form = UploadForm(self)
-        if form.validate_on_submit():
-            filename = op.join(directory,
-                               secure_filename(form.upload.data.filename))
+        if not self.is_accessible_path(path):
+            flash(gettext('Permission denied.'))
+            return redirect(self._get_dir_url('.index'))
 
-            if op.exists(filename):
-                flash(gettext('File "%(name)s" already exists.', name=filename),
-                      'error')
-            else:
-                try:
-                    self.save_file(filename, form.upload.data)
-                    self.on_file_upload(directory, path, filename)
-                    return redirect(self._get_dir_url('.index', path))
-                except Exception, ex:
-                    flash(gettext('Failed to save file: %(error)s', error=ex))
+        form = self.upload_form(self)
+        if helpers.validate_form_on_submit(form):
+            try:
+                self._save_form_files(directory, path, form)
+                return redirect(self._get_dir_url('.index', path))
+            except Exception as ex:
+                flash(gettext('Failed to save file: %(error)s', error=ex))
 
         return self.render(self.upload_template, form=form)
 
+    @expose('/download/<path:path>')
+    def download(self, path=None):
+        """
+            Download view method.
+
+            :param path:
+                File path.
+        """
+        if not self.can_download:
+            abort(404)
+
+        base_path, directory, path = self._normalize_path(path)
+
+        # backward compatibility with base_url
+        base_url = self.get_base_url()
+        if base_url:
+            base_url = urljoin(url_for('.index'), base_url)
+            return redirect(urljoin(base_url, path))
+
+        return send_file(directory)
+
     @expose('/mkdir/', methods=('GET', 'POST'))
     @expose('/mkdir/<path:path>', methods=('GET', 'POST'))
     def mkdir(self, path=None):
@@ -511,15 +554,19 @@ class FileAdmin(BaseView, ActionsMixin):
             flash(gettext('Directory creation is disabled.'), 'error')
             return redirect(dir_url)
 
-        form = NameForm(request.form)
+        if not self.is_accessible_path(path):
+            flash(gettext('Permission denied.'))
+            return redirect(self._get_dir_url('.index'))
 
-        if form.validate_on_submit():
+        form = NameForm(helpers.get_form_data())
+
+        if helpers.validate_form_on_submit(form):
             try:
                 os.mkdir(op.join(directory, form.name.data))
                 self.on_mkdir(directory, form.name.data)
                 return redirect(dir_url)
-            except Exception, ex:
-                flash(gettext('Failed to create directory: %(error)s', ex), 'error')
+            except Exception as ex:
+                flash(gettext('Failed to create directory: %(error)s', error=ex), 'error')
 
         return self.render(self.mkdir_template,
                            form=form,
@@ -544,6 +591,10 @@ class FileAdmin(BaseView, ActionsMixin):
             flash(gettext('Deletion is disabled.'))
             return redirect(return_url)
 
+        if not self.is_accessible_path(path):
+            flash(gettext('Permission denied.'))
+            return redirect(self._get_dir_url('.index'))
+
         if op.isdir(full_path):
             if not self.can_delete_dirs:
                 flash(gettext('Directory deletion is disabled.'))
@@ -553,14 +604,14 @@ class FileAdmin(BaseView, ActionsMixin):
                 shutil.rmtree(full_path)
                 self.on_directory_delete(full_path, path)
                 flash(gettext('Directory "%s" was successfully deleted.' % path))
-            except Exception, ex:
+            except Exception as ex:
                 flash(gettext('Failed to delete directory: %(error)s', error=ex), 'error')
         else:
             try:
                 os.remove(full_path)
                 self.on_file_delete(full_path, path)
                 flash(gettext('File "%(name)s" was successfully deleted.', name=path))
-            except Exception, ex:
+            except Exception as ex:
                 flash(gettext('Failed to delete file: %(name)s', name=ex), 'error')
 
         return redirect(return_url)
@@ -583,12 +634,16 @@ class FileAdmin(BaseView, ActionsMixin):
             flash(gettext('Renaming is disabled.'))
             return redirect(return_url)
 
+        if not self.is_accessible_path(path):
+            flash(gettext('Permission denied.'))
+            return redirect(self._get_dir_url('.index'))
+
         if not op.exists(full_path):
             flash(gettext('Path does not exist.'))
             return redirect(return_url)
 
-        form = NameForm(request.form, name=op.basename(path))
-        if form.validate_on_submit():
+        form = NameForm(helpers.get_form_data(), name=op.basename(path))
+        if helpers.validate_form_on_submit(form):
             try:
                 dir_base = op.dirname(full_path)
                 filename = secure_filename(form.name.data)
@@ -598,7 +653,7 @@ class FileAdmin(BaseView, ActionsMixin):
                 flash(gettext('Successfully renamed "%(src)s" to "%(dst)s"',
                       src=op.basename(path),
                       dst=filename))
-            except Exception, ex:
+            except Exception as ex:
                 flash(gettext('Failed to rename: %(error)s', error=ex), 'error')
 
             return redirect(return_url)
@@ -624,13 +679,18 @@ class FileAdmin(BaseView, ActionsMixin):
         path = path[0]
 
         base_path, full_path, path = self._normalize_path(path)
+
+        if not self.is_accessible_path(path) or not self.is_file_editable(path):
+            flash(gettext('Permission denied.'))
+            return redirect(self._get_dir_url('.index'))
+
         dir_url = self._get_dir_url('.index', os.path.dirname(path))
         next_url = next_url or dir_url
 
-        form = EditForm()
+        form = EditForm(helpers.get_form_data())
         error = False
 
-        if request.method == 'POST':
+        if helpers.validate_form_on_submit(form):
             form.process(request.form, content='')
             if form.validate():
                 try:
@@ -655,7 +715,7 @@ class FileAdmin(BaseView, ActionsMixin):
                 error = True
             else:
                 try:
-                    content.decode('utf8')
+                    content = content.decode('utf8')
                 except UnicodeDecodeError:
                     flash(gettext("Cannot edit %(name)s.", name=path), 'error')
                     error = True
@@ -677,14 +737,19 @@ class FileAdmin(BaseView, ActionsMixin):
             lazy_gettext('Delete'),
             lazy_gettext('Are you sure you want to delete these files?'))
     def action_delete(self, items):
+        if not self.can_delete:
+            flash(gettext('File deletion is disabled.'), 'error')
+            return
+
         for path in items:
             base_path, full_path, path = self._normalize_path(path)
 
-            try:
-                os.remove(full_path)
-                flash(gettext('File "%(name)s" was successfully deleted.', name=path))
-            except Exception, ex:
-                flash(gettext('Failed to delete file: %(name)s', name=ex), 'error')
+            if self.is_accessible_path(path):
+                try:
+                    os.remove(full_path)
+                    flash(gettext('File "%(name)s" was successfully deleted.', name=path))
+                except Exception as ex:
+                    flash(gettext('Failed to delete file: %(name)s', name=ex), 'error')
 
     @action('edit', lazy_gettext('Edit'))
     def action_edit(self, items):
