@@ -5,16 +5,19 @@ from bson import ObjectId
 from bson.errors import InvalidId
 
 from flask import flash
-from jinja2 import contextfunction
 
+from flask.ext.admin._compat import string_types
 from flask.ext.admin.babel import gettext, ngettext, lazy_gettext
 from flask.ext.admin.model import BaseModelView
-from flask.ext.admin.model.helpers import get_default_order
 from flask.ext.admin.actions import action
+from flask.ext.admin.helpers import get_form_data
 
 from .filters import BasePyMongoFilter
 from .tools import parse_like_term
 
+# Set up logger
+log = logging.getLogger("flask-admin.pymongo")
+
 
 class ModelView(BaseModelView):
     """
@@ -93,7 +96,7 @@ class ModelView(BaseModelView):
         """
         if self.column_searchable_list:
             for p in self.column_searchable_list:
-                if not isinstance(p, basestring):
+                if not isinstance(p, string_types):
                     raise ValueError('Expected string')
 
                 # TODO: Validation?
@@ -207,7 +210,7 @@ class ModelView(BaseModelView):
         if sort_column:
             sort_by = [(sort_column, pymongo.DESCENDING if sort_desc else pymongo.ASCENDING)]
         else:
-            order = get_default_order(self)
+            order = self._get_default_order()
 
             if order:
                 sort_by = [(order[0], pymongo.DESCENDING if order[1] else pymongo.ASCENDING)]
@@ -244,7 +247,7 @@ class ModelView(BaseModelView):
         """
             Create edit form from the MongoDB document
         """
-        return self._edit_form_class(**obj)
+        return self._edit_form_class(get_form_data(), **obj)
 
     def create_model(self, form):
         """
@@ -255,12 +258,12 @@ class ModelView(BaseModelView):
         """
         try:
             model = form.data
-            self.on_model_change(form, model)
+            self._on_model_change(form, model, True)
             self.coll.insert(model)
-        except Exception, ex:
+        except Exception as ex:
             flash(gettext('Failed to create model. %(error)s', error=str(ex)),
                   'error')
-            logging.exception('Failed to create model')
+            log.exception('Failed to create model')
             return False
         else:
             self.after_model_change(form, model, True)
@@ -278,14 +281,14 @@ class ModelView(BaseModelView):
         """
         try:
             model.update(form.data)
-            self.on_model_change(form, model)
+            self._on_model_change(form, model, False)
 
             pk = self.get_pk_value(model)
             self.coll.update({'_id': pk}, model)
-        except Exception, ex:
+        except Exception as ex:
             flash(gettext('Failed to update model. %(error)s', error=str(ex)),
                   'error')
-            logging.exception('Failed to update model')
+            log.exception('Failed to update model')
             return False
         else:
             self.after_model_change(form, model, False)
@@ -308,10 +311,10 @@ class ModelView(BaseModelView):
             self.on_model_delete(model)
             self.coll.remove({'_id': pk})
             return True
-        except Exception, ex:
+        except Exception as ex:
             flash(gettext('Failed to delete model. %(error)s', error=str(ex)),
                   'error')
-            logging.exception('Failed to delete model')
+            log.exception('Failed to delete model')
             return False
 
     # Default model actions
@@ -338,6 +341,5 @@ class ModelView(BaseModelView):
                            '%(count)s models were successfully deleted.',
                            count,
                            count=count))
-        except Exception, ex:
-            flash(gettext('Failed to delete models. %(error)s', error=str(ex)),
-                'error')
+        except Exception as ex:
+            flash(gettext('Failed to delete models. %(error)s', error=str(ex)), 'error')
