@@ -84,12 +84,10 @@ class AdminModelConverter(ModelConverterBase):
             kwargs['query_factory'] = lambda: self.session.query(remote_model)
 
         if 'widget' not in kwargs:
-            if prop.direction.name == 'MANYTOONE':
+            if multiple:
+                kwargs['widget'] = form.Select2Widget(multiple=True)
+            else:
                 kwargs['widget'] = form.Select2Widget()
-            elif prop.direction.name == 'ONETOMANY':
-                kwargs['widget'] = form.Select2Widget(multiple=True)
-            elif prop.direction.name == 'MANYTOMANY':
-                kwargs['widget'] = form.Select2Widget(multiple=True)
 
         if multiple:
             return QuerySelectMultipleField(**kwargs)
@@ -97,6 +95,11 @@ class AdminModelConverter(ModelConverterBase):
             return QuerySelectField(**kwargs)
 
     def _convert_relation(self, prop, kwargs):
+        # Check if relation is specified
+        form_columns = getattr(self.view, 'form_columns', None)
+        if form_columns and prop.key not in form_columns:
+            return None
+
         remote_model = prop.mapper.class_
         column = prop.local_remote_pairs[0][0]
 
@@ -108,9 +111,9 @@ class AdminModelConverter(ModelConverterBase):
         kwargs['label'] = self._get_label(prop.key, kwargs)
         kwargs['description'] = self._get_description(prop.key, kwargs)
 
-        if column.nullable:
+        if column.nullable or prop.direction.name != 'MANYTOONE':
             kwargs['validators'].append(validators.Optional())
-        elif prop.direction.name != 'MANYTOMANY':
+        else:
             kwargs['validators'].append(validators.InputRequired())
 
         # Contribute model-related parameters
@@ -122,7 +125,7 @@ class AdminModelConverter(ModelConverterBase):
         if override:
             return override(**kwargs)
 
-        if prop.direction.name == 'MANYTOONE':
+        if prop.direction.name == 'MANYTOONE' or not prop.uselist:
             return self._model_select_field(prop, False, remote_model, **kwargs)
         elif prop.direction.name == 'ONETOMANY':
             return self._model_select_field(prop, True, remote_model, **kwargs)
@@ -158,8 +161,10 @@ class AdminModelConverter(ModelConverterBase):
                     # Grab column
                     column = prop.columns[0]
 
-                # Do not display foreign keys - use relations
-                if column.foreign_keys:
+                form_columns = getattr(self.view, 'form_columns', None) or ()
+
+                # Do not display foreign keys - use relations, except when explicitly instructed
+                if column.foreign_keys and prop.key not in form_columns:
                     return None
 
                 # Only display "real" columns
@@ -174,11 +179,6 @@ class AdminModelConverter(ModelConverterBase):
                         return fields.HiddenField()
                     else:
                         # By default, don't show primary keys either
-                        form_columns = getattr(self.view, 'form_columns', None)
-
-                        if form_columns is None:
-                            return None
-
                         # If PK is not explicitly allowed, ignore it
                         if prop.key not in form_columns:
                             return None
@@ -267,6 +267,11 @@ class AdminModelConverter(ModelConverterBase):
             field_args['choices'] = [(f, f) for f in column.type.enums]
             return form.Select2Field(**field_args)
 
+        if column.nullable:
+            filters = field_args.get('filters', [])
+            filters.append(lambda x: x or None)
+            field_args['filters'] = filters
+
         self._string_common(column=column, field_args=field_args, **extra)
         return fields.TextField(**field_args)
 
@@ -545,7 +550,7 @@ class InlineModelConverter(InlineModelConverterBase):
         reverse_prop = None
 
         for prop in target_mapper.iterate_properties:
-            if hasattr(prop, 'direction') and prop.direction.name == 'MANYTOONE':
+            if hasattr(prop, 'direction'):
                 if issubclass(model, prop.mapper.class_):
                     reverse_prop = prop
                     break
@@ -556,7 +561,7 @@ class InlineModelConverter(InlineModelConverterBase):
         forward_prop = None
 
         for prop in mapper.iterate_properties:
-            if hasattr(prop, 'direction') and prop.direction.name == 'ONETOMANY':
+            if hasattr(prop, 'direction'):
                 if prop.mapper.class_ == target_mapper.class_:
                     forward_prop = prop
                     break
