@@ -1,15 +1,27 @@
 import warnings
+import re
 
-from flask import request, url_for, redirect, flash
+from flask import request, url_for, redirect, flash, abort, json, Response
 
 from jinja2 import contextfunction
 
 from flask.ext.admin.babel import gettext
 
 from flask.ext.admin.base import BaseView, expose
-from flask.ext.admin.tools import rec_getattr, ObsoleteAttr
+from flask.ext.admin.form import BaseForm, FormOpts, rules
 from flask.ext.admin.model import filters, typefmt
 from flask.ext.admin.actions import ActionsMixin
+from flask.ext.admin.helpers import get_form_data, validate_form_on_submit, get_redirect_target
+from flask.ext.admin.tools import rec_getattr
+from flask.ext.admin._backwards import ObsoleteAttr
+from flask.ext.admin._compat import iteritems, OrderedDict
+from .helpers import prettify_name, get_mdict_item_or_list
+from .ajax import AjaxModelLoader
+
+
+# Used to generate filter query string name
+filter_char_re = re.compile('[^a-z0-9 ]')
+filter_compact_re = re.compile(' +')
 
 
 class BaseModelView(BaseView, ActionsMixin):
@@ -83,6 +95,18 @@ class BaseModelView(BaseView, ActionsMixin):
             class MyModelView(BaseModelView):
                 column_formatters = dict(price=lambda v, c, m, p: m.price*2)
 
+        or using Jinja2 `macro` in template::
+
+            from flask.ext.admin.model.template import macro
+
+            class MyModelView(BaseModelView):
+                column_formatters = dict(price=macro('render_price'))
+
+            # in template
+            {% macro render_price(model, column) %}
+                {{ model.price * 2 }}
+            {% endmacro %}
+
         The Callback function has the prototype::
 
             def formatter(view, context, model, name):
@@ -110,9 +134,10 @@ class BaseModelView(BaseView, ActionsMixin):
         If you want to display `NULL` instead of an empty string, you can do
         something like this::
 
-            from flask.ext.admin import typefmt
+            from flask.ext.admin.model import typefmt
 
-            MY_DEFAULT_FORMATTERS = dict(typefmt.BASE_FORMATTERS).extend({
+            MY_DEFAULT_FORMATTERS = dict(typefmt.BASE_FORMATTERS)
+            MY_DEFAULT_FORMATTERS.update({
                     type(None): typefmt.null_formatter
                 })
 
@@ -233,6 +258,15 @@ class BaseModelView(BaseView, ActionsMixin):
                 column_filters = ('user', 'email')
     """
 
+    named_filter_urls = False
+    """
+        Set to True to use human-readable names for filters in URL parameters.
+
+        False by default so as to be robust across translations.
+
+        Changing this parameter will break any existing URLs that have filters.
+    """
+
     column_display_pk = ObsoleteAttr('column_display_pk',
                                      'list_display_pk',
                                      False)
@@ -243,16 +277,34 @@ class BaseModelView(BaseView, ActionsMixin):
     form = None
     """
         Form class. Override if you want to use custom form for your model.
+        Will completely disable form scaffolding functionality.
 
         For example::
 
-            class MyForm(wtf.Form):
-                pass
+            class MyForm(Form):
+                name = TextField('Name')
 
             class MyModelView(BaseModelView):
                 form = MyForm
     """
 
+    form_base_class = BaseForm
+    """
+        Base form class. Will be used by form scaffolding function when creating model form.
+
+        Useful if you want to have custom contructor or override some fields.
+
+        Example::
+
+            class MyBaseForm(Form):
+                def do_something(self):
+                    pass
+
+            class MyModelView(BaseModelView):
+                form_base_class = MyBaseForm
+
+    """
+
     form_args = None
     """
         Dictionary of form field arguments. Refer to WTForms documentation for
@@ -262,7 +314,7 @@ class BaseModelView(BaseView, ActionsMixin):
 
             class MyModelView(BaseModelView):
                 form_args = dict(
-                    name=dict(label='First Name', validators=[wtf.required()])
+                    name=dict(label='First Name', validators=[required()])
                 )
     """
 
@@ -315,6 +367,95 @@ class BaseModelView(BaseView, ActionsMixin):
                 }
     """
 
+    form_extra_fields = None
+    """
+        Dictionary of additional fields.
+
+        Example::
+
+            class MyModelView(BaseModelView):
+                form_extra_fields = {
+                    password: PasswordField('Password')
+                }
+
+        You can control order of form fields using ``form_columns`` property. For example::
+
+            class MyModelView(BaseModelView):
+                form_columns = ('name', 'email', 'password', 'secret')
+
+                form_extra_fields = {
+                    password: PasswordField('Password')
+                }
+
+        In this case, password field will be put between email and secret fields that are autogenerated.
+    """
+
+    form_ajax_refs = None
+    """
+        Use AJAX for foreign key model loading.
+
+        Should contain dictionary, where key is field name and value is either a dictionary which
+        configures AJAX lookups or backend-specific `AjaxModelLoader` class instance.
+
+        For example, it can look like::
+
+            class MyModelView(BaseModelView):
+                form_ajax_refs = {
+                    'user': {
+                        'fields': ('first_name', 'last_name', 'email')
+                        'page_size': 10
+                    }
+                }
+
+        Or with SQLAlchemy backend like this::
+
+            class MyModelView(BaseModelView):
+                form_ajax_refs = {
+                    'user': QueryAjaxModelLoader('user', db.session, User, fields=['email'], page_size=10)
+                }
+
+        If you need custom loading functionality, you can implement your custom loading behavior
+        in your `AjaxModelLoader` class.
+    """
+
+    form_rules = None
+    """
+        List of rendering rules for model creation form.
+
+        This property changed default form rendering behavior and makes possible to rearrange order
+        of rendered fields, add some text between fields, group them, etc. If not set, will use
+        default Flask-Admin form rendering logic.
+
+        Here's simple example which illustrates how to use::
+
+            from flask.ext.admin.form import rules
+
+            class MyModelView(ModelView):
+                form_rules = [
+                    # Define field set with header text and four fields
+                    rules.FieldSet(('first_name', 'last_name', 'email', 'phone'), 'User'),
+                    # ... and it is just shortcut for:
+                    rules.Header('User'),
+                    rules.Field('first_name'),
+                    rules.Field('last_name'),
+                    # ...
+                    # It is possible to create custom rule blocks:
+                    MyBlock('Hello World'),
+                    # It is possible to call macros from current context
+                    rules.Macro('my_macro', foobar='baz')
+                ]
+    """
+
+    form_edit_rules = None
+    """
+        Customized rules for the edit form. Override `form_rules` if present.
+    """
+
+    form_create_rules = None
+    """
+        Customized rules for the create form. Override `form_rules` if present.
+    """
+
     # Actions
     action_disallowed_list = ObsoleteAttr('action_disallowed_list',
                                           'disallowed_actions',
@@ -350,11 +491,14 @@ class BaseModelView(BaseView, ActionsMixin):
                 'userview'
             :param url:
                 Base URL. If not provided, will use endpoint as a URL.
+            :param debug:
+                Enable debugging mode. Won't catch exceptions on model
+                save failures.
         """
 
         # If name not provided, it is model name
         if name is None:
-            name = '%s' % self._prettify_name(model.__name__)
+            name = '%s' % self._prettify_class_name(model.__name__)
 
         # If endpoint not provided, it is model name + 'view'
         if endpoint is None:
@@ -384,12 +528,14 @@ class BaseModelView(BaseView, ActionsMixin):
             self.column_labels = {}
 
         # Forms
-        self._create_form_class = self.get_create_form()
-        self._edit_form_class = self.get_edit_form()
+        self._form_ajax_refs = self._process_ajax_references()
 
         if self.form_widget_args is None:
             self.form_widget_args = {}
 
+        self._create_form_class = self.get_create_form()
+        self._edit_form_class = self.get_edit_form()
+
         # Search
         self._search_supported = self.init_search()
 
@@ -412,26 +558,47 @@ class BaseModelView(BaseView, ActionsMixin):
         if self.column_descriptions is None:
             self.column_descriptions = dict()
 
+        # Group filters by field name
         if self._filters:
-            self._filter_groups = []
-            self._filter_dict = dict()
+            self._filter_groups = OrderedDict()
+            self._filter_args = {}
 
-            for i, n in enumerate(self._filters):
-                if n.name not in self._filter_dict:
-                    group = []
-                    self._filter_dict[n.name] = group
-                    self._filter_groups.append((n.name, group))
-                else:
-                    group = self._filter_dict[n.name]
+            for i, flt in enumerate(self._filters):
+                if flt.name not in self._filter_groups:
+                    self._filter_groups[flt.name] = []
 
-                group.append((i, n.operation()))
+                self._filter_groups[flt.name].append({
+                    'index': i,
+                    'arg': self.get_filter_arg(i, flt),
+                    'operation': flt.operation(),
+                    'options': flt.get_options(self) or None,
+                    'type': flt.data_type
+                })
 
-            self._filter_types = dict((i, f.data_type)
-                                      for i, f in enumerate(self._filters)
-                                      if f.data_type)
+                self._filter_args[self.get_filter_arg(i, flt)] = (i, flt)
         else:
             self._filter_groups = None
-            self._filter_types = None
+            self._filter_args = None
+
+        # Form rendering rules
+        if self.form_create_rules:
+            self._form_create_rules = rules.RuleSet(self, self.form_create_rules)
+        else:
+            self._form_create_rules = None
+
+        if self.form_edit_rules:
+            self._form_edit_rules = rules.RuleSet(self, self.form_edit_rules)
+        else:
+            self._form_edit_rules = None
+
+        if self.form_rules:
+            form_rules = rules.RuleSet(self, self.form_rules)
+
+            if not self._form_create_rules:
+                self._form_create_rules = form_rules
+
+            if not self._form_edit_rules:
+                self._form_edit_rules = form_rules
 
     # Primary key
     def get_pk_value(self, model):
@@ -463,7 +630,7 @@ class BaseModelView(BaseView, ActionsMixin):
         if self.column_labels and field in self.column_labels:
             return self.column_labels[field]
         else:
-            return self.prettify_name(field)
+            return self._prettify_name(field)
 
     def get_list_columns(self):
         """
@@ -520,6 +687,7 @@ class BaseModelView(BaseView, ActionsMixin):
         """
         return False
 
+    # Filter helpers
     def scaffold_filters(self, name):
         """
             Generate filter object for the given name
@@ -564,6 +732,27 @@ class BaseModelView(BaseView, ActionsMixin):
         else:
             return None
 
+    def get_filter_arg(self, index, flt):
+        """
+            Given a filter `flt`, return a unique name for that filter in
+            this view.
+
+            Does not include the `flt[n]_` portion of the filter name.
+
+            :param index:
+                Filter index in _filters array
+            :param flt:
+                Filter instance
+        """
+        if self.named_filter_urls:
+            name = ('%s %s' % (flt.name, flt.operation())).lower()
+            name = filter_char_re.sub('', name)
+            name = filter_compact_re.sub('_', name)
+            return name
+        else:
+            return str(index)
+
+    # Form helpers
     def scaffold_form(self):
         """
             Create `form.BaseForm` inherited class from the model. Must be
@@ -607,7 +796,7 @@ class BaseModelView(BaseView, ActionsMixin):
 
             Override to implement custom behavior.
         """
-        return self._create_form_class(obj=obj)
+        return self._create_form_class(get_form_data(), obj=obj)
 
     def edit_form(self, obj=None):
         """
@@ -615,7 +804,7 @@ class BaseModelView(BaseView, ActionsMixin):
 
             Override to implement custom behavior.
         """
-        return self._edit_form_class(obj=obj)
+        return self._edit_form_class(get_form_data(), obj=obj)
 
     # Helpers
     def is_sortable(self, name):
@@ -636,6 +825,18 @@ class BaseModelView(BaseView, ActionsMixin):
 
         return self._list_columns[idx]
 
+    def _get_default_order(self):
+        """
+            Return default sort order
+        """
+        if self.column_default_sort:
+            if isinstance(self.column_default_sort, tuple):
+                return self.column_default_sort
+            else:
+                return self.column_default_sort, False
+
+        return None
+
     # Database-related API
     def get_list(self, page, sort_field, sort_desc, search, filters):
         """
@@ -669,7 +870,7 @@ class BaseModelView(BaseView, ActionsMixin):
         raise NotImplemented('Please implement get_one method')
 
     # Model event handlers
-    def on_model_change(self, form, model):
+    def on_model_change(self, form, model, is_created):
         """
             Perform some actions after a model is created or updated.
 
@@ -682,9 +883,24 @@ class BaseModelView(BaseView, ActionsMixin):
                 Form used to create/update model
             :param model:
                 Model that will be created/updated
+            :param is_created:
+                Will be set to True if model was created and to False if edited
         """
         pass
 
+    def _on_model_change(self, form, model, is_created):
+        """
+            Compatibility helper.
+        """
+        try:
+            self.on_model_change(form, model, is_created)
+        except TypeError:
+            msg = ('%s.on_model_change() now accepts third ' +
+                   'parameter is_created. Please update your code') % self.model
+            warnings.warn(msg)
+
+            self.on_model_change(form, model)
+
     def after_model_change(self, form, model, is_created):
         """
             Perform some actions after a model was created or updated and
@@ -756,7 +972,7 @@ class BaseModelView(BaseView, ActionsMixin):
         raise NotImplemented()
 
     # Various helpers
-    def prettify_name(self, name):
+    def _prettify_name(self, name):
         """
             Prettify pythonic variable name.
 
@@ -765,10 +981,39 @@ class BaseModelView(BaseView, ActionsMixin):
             :param name:
                 Name to prettify
         """
-        return name.replace('_', ' ').title()
+        return prettify_name(name)
 
-    # URL generation helper
-    def _get_extra_args(self):
+    def get_empty_list_message(self):
+        return gettext('There are no items in the table.')
+
+    # URL generation helpers
+    def _get_list_filter_args(self):
+        if self._filters:
+            filters = []
+
+            for n in request.args:
+                if not n.startswith('flt'):
+                    continue
+
+                if '_' not in n:
+                    continue
+
+                pos, key = n[3:].split('_', 1)
+
+                if key in self._filter_args:
+                    idx, flt = self._filter_args[key]
+
+                    value = request.args[n]
+
+                    if flt.validate(value):
+                        filters.append((pos, (idx, flt.clean(value))))
+
+            # Sort filters
+            return [v[1] for v in sorted(filters, key=lambda n: n[0])]
+
+        return None
+
+    def _get_list_extra_args(self):
         """
             Return arguments from query string.
         """
@@ -776,34 +1021,7 @@ class BaseModelView(BaseView, ActionsMixin):
         sort = request.args.get('sort', None, type=int)
         sort_desc = request.args.get('desc', None, type=int)
         search = request.args.get('search', None)
-
-        # Gather filters
-        if self._filters:
-            sfilters = []
-
-            for n in request.args:
-                if n.startswith('flt'):
-                    ofs = n.find('_')
-                    if ofs == -1:
-                        continue
-
-                    try:
-                        pos = int(n[3:ofs])
-                        idx = int(n[ofs + 1:])
-                    except ValueError:
-                        continue
-
-                    if idx >= 0 and idx < len(self._filters):
-                        flt = self._filters[idx]
-
-                        value = request.args[n]
-
-                        if flt.validate(value):
-                            sfilters.append((pos, (idx, flt.clean(value))))
-
-            filters = [v[1] for v in sorted(sfilters, key=lambda n: n[0])]
-        else:
-            filters = None
+        filters = self._get_list_filter_args()
 
         return page, sort, sort_desc, search, filters
 
@@ -835,9 +1053,11 @@ class BaseModelView(BaseView, ActionsMixin):
         kwargs = dict(page=page, sort=sort, desc=sort_desc, search=search)
 
         if filters:
-            for i, flt in enumerate(filters):
-                key = 'flt%d_%d' % (i, flt[0])
-                kwargs[key] = flt[1]
+            for i, pair in enumerate(filters):
+                idx, value = pair
+
+                key = 'flt%d_%s' % (i, self.get_filter_arg(idx, self._filters[idx]))
+                kwargs[key] = value
 
         return url_for(view, **kwargs)
 
@@ -871,17 +1091,9 @@ class BaseModelView(BaseView, ActionsMixin):
         """
         column_fmt = self.column_formatters.get(name)
         if column_fmt is not None:
-            try:
-                return column_fmt(self, context, model, name)
-            except TypeError:
-                warnings.warn('Column formatter prototype was changed to accept view as first input parameter.\n' +
-                              'Please update %s %s formatter to accept 4 parameters.' % (self.name, name),
-                              stacklevel=2)
-                self.column_formatters[name] = lambda _, c, m, n: column_fmt(c, m, n)
-
-                return column_fmt(context, model, name)
-
-        value = self._get_field_value(model, name)
+            value = column_fmt(self, context, model, name)
+        else:
+            value = self._get_field_value(model, name)
 
         choices_map = self._column_choices_map.get(name, {})
         if choices_map:
@@ -889,18 +1101,35 @@ class BaseModelView(BaseView, ActionsMixin):
 
         type_fmt = self.column_type_formatters.get(type(value))
         if type_fmt is not None:
-            try:
-                value = type_fmt(self, value)
-            except TypeError:
-                warnings.warn('Type formatter prototype was changed to accept view as first input parameter.\n' +
-                              'Please update %s %s formatter to accept 2 parameters.' % (self.name, type(value)),
-                              stacklevel=2)
-                self.column_type_formatters[type(value)] = lambda _, value: type_fmt(value)
-
-                value = type_fmt(value)
+            value = type_fmt(self, value)
 
         return value
 
+    # AJAX references
+    def _process_ajax_references(self):
+        """
+            Process `form_ajax_refs` and generate model loaders that
+            will be used by the `ajax_lookup` view.
+        """
+        result = {}
+
+        if self.form_ajax_refs:
+            for name, options in iteritems(self.form_ajax_refs):
+                if isinstance(options, dict):
+                    result[name] = self._create_ajax_loader(name, options)
+                elif isinstance(options, AjaxModelLoader):
+                    result[name] = options
+                else:
+                    raise ValueError('%s.form_ajax_refs can not handle %s types' % (self, type(options)))
+
+        return result
+
+    def _create_ajax_loader(self, name, options):
+        """
+            Model backend will override this to implement AJAX model loading.
+        """
+        raise NotImplemented()
+
     # Views
     @expose('/')
     def index_view(self):
@@ -908,7 +1137,7 @@ class BaseModelView(BaseView, ActionsMixin):
             List view
         """
         # Grab parameters from URL
-        page, sort_idx, sort_desc, search, filters = self._get_extra_args()
+        page, sort_idx, sort_desc, search, filters = self._get_list_extra_args()
 
         # Map column index to column name
         sort_column = self._get_column_by_idx(sort_idx)
@@ -920,22 +1149,10 @@ class BaseModelView(BaseView, ActionsMixin):
                                     search, filters)
 
         # Calculate number of pages
-        num_pages = count / self.page_size
+        num_pages = count // self.page_size
         if count % self.page_size != 0:
             num_pages += 1
 
-        # Pregenerate filters
-        if self._filters:
-            filters_data = dict()
-
-            for idx, f in enumerate(self._filters):
-                flt_data = f.get_options(self)
-
-                if flt_data:
-                    filters_data[idx] = flt_data
-        else:
-            filters_data = None
-
         # Various URL generation helpers
         def pager_url(p):
             # Do not add page number if it is first page
@@ -991,28 +1208,25 @@ class BaseModelView(BaseView, ActionsMixin):
                                # Filters
                                filters=self._filters,
                                filter_groups=self._filter_groups,
-                               filter_types=self._filter_types,
-                               filter_data=filters_data,
                                active_filters=filters,
 
                                # Actions
                                actions=actions,
-                               actions_confirmation=actions_confirmation
-                               )
+                               actions_confirmation=actions_confirmation)
 
     @expose('/new/', methods=('GET', 'POST'))
     def create_view(self):
         """
             Create model view
         """
-        return_url = request.args.get('url') or url_for('.index_view')
+        return_url = get_redirect_target() or url_for('.index_view')
 
         if not self.can_create:
             return redirect(return_url)
 
         form = self.create_form()
 
-        if form.validate_on_submit():
+        if validate_form_on_submit(form):
             if self.create_model(form):
                 if '_add_another' in request.form:
                     flash(gettext('Model was successfully created.'))
@@ -1020,9 +1234,12 @@ class BaseModelView(BaseView, ActionsMixin):
                 else:
                     return redirect(return_url)
 
+        form_opts = FormOpts(widget_args=self.form_widget_args,
+                             form_rules=self._form_create_rules)
+
         return self.render(self.create_template,
                            form=form,
-                           form_widget_args=self.form_widget_args,
+                           form_opts=form_opts,
                            return_url=return_url)
 
     @expose('/edit/', methods=('GET', 'POST'))
@@ -1030,12 +1247,12 @@ class BaseModelView(BaseView, ActionsMixin):
         """
             Edit model view
         """
-        return_url = request.args.get('url') or url_for('.index_view')
+        return_url = get_redirect_target() or url_for('.index_view')
 
         if not self.can_edit:
             return redirect(return_url)
 
-        id = request.args.get('id')
+        id = get_mdict_item_or_list(request.args, 'id')
         if id is None:
             return redirect(return_url)
 
@@ -1046,14 +1263,21 @@ class BaseModelView(BaseView, ActionsMixin):
 
         form = self.edit_form(obj=model)
 
-        if form.validate_on_submit():
+        if validate_form_on_submit(form):
             if self.update_model(form, model):
-                return redirect(return_url)
+                if '_continue_editing' in request.form:
+                    flash(gettext('Model was successfully saved.'))
+                    return redirect(request.url)
+                else:
+                    return redirect(return_url)
+
+        form_opts = FormOpts(widget_args=self.form_widget_args,
+                             form_rules=self._form_edit_rules)
 
         return self.render(self.edit_template,
                            model=model,
                            form=form,
-                           form_widget_args=self.form_widget_args,
+                           form_opts=form_opts,
                            return_url=return_url)
 
     @expose('/delete/', methods=('POST',))
@@ -1061,13 +1285,13 @@ class BaseModelView(BaseView, ActionsMixin):
         """
             Delete model view. Only POST method is allowed.
         """
-        return_url = request.args.get('url') or url_for('.index_view')
+        return_url = get_redirect_target() or url_for('.index_view')
 
         # TODO: Use post
         if not self.can_delete:
             return redirect(return_url)
 
-        id = request.args.get('id')
+        id = get_mdict_item_or_list(request.args, 'id')
         if id is None:
             return redirect(return_url)
 
@@ -1084,3 +1308,18 @@ class BaseModelView(BaseView, ActionsMixin):
             Mass-model action view.
         """
         return self.handle_action()
+
+    @expose('/ajax/lookup/')
+    def ajax_lookup(self):
+        name = request.args.get('name')
+        query = request.args.get('query')
+        offset = request.args.get('offset', type=int)
+        limit = request.args.get('limit', 10, type=int)
+
+        loader = self._form_ajax_refs.get(name)
+
+        if not loader:
+            abort(404)
+
+        data = [loader.format(m) for m in loader.get_list(query, offset, limit)]
+        return Response(json.dumps(data), mimetype='application/json')
