@@ -76,7 +76,7 @@ class InlineFieldList(FieldList):
         _fake = type(str('_fake'), (object, ), {})
 
         output = []
-        for field, data in itertools.izip(self.entries, candidates):
+        for field, data in zip(self.entries, candidates):
             if not self.should_delete(field):
                 fake_obj = _fake()
                 fake_obj.data = data
@@ -102,10 +102,11 @@ class InlineModelFormField(FormField):
     """
     widget = InlineFormWidget()
 
-    def __init__(self, form_class, pk, **kwargs):
+    def __init__(self, form_class, pk, form_opts=None, **kwargs):
         super(InlineModelFormField, self).__init__(form_class, **kwargs)
 
         self._pk = pk
+        self.form_opts = form_opts
 
     def get_pk(self):
         return getattr(self.form, self._pk).data
