@@ -1,3 +1,4 @@
+from flask.ext.admin._compat import text_type
 from flask.ext.admin.babel import lazy_gettext
 
 
@@ -29,8 +30,13 @@ class BaseFilter(object):
             :param view:
                 Associated administrative view class.
         """
-        if self.options:
-            return [(v, unicode(n)) for v, n in self.options]
+        options = self.options
+
+        if options:
+            if callable(options):
+                options = options()
+
+            return [(v, text_type(n)) for v, n in options]
 
         return None
 
@@ -82,12 +88,12 @@ class BaseBooleanFilter(BaseFilter):
     """
     def __init__(self, name, options=None, data_type=None):
         super(BaseBooleanFilter, self).__init__(name,
-                                                (('1', lazy_gettext('Yes')),
-                                                 ('0', lazy_gettext('No'))),
+                                                (('1', lazy_gettext(u'Yes')),
+                                                 ('0', lazy_gettext(u'No'))),
                                                 data_type)
 
     def validate(self, value):
-        return value == '0' or value == '1'
+        return value in ('0', '1')
 
 
 class BaseDateFilter(BaseFilter):
@@ -122,7 +128,7 @@ def convert(*args):
     """
         Decorator for field to filter conversion routine.
 
-        See :mod:`flask.ext.admin.contrib.sqlamodel.filters` for usage example.
+        See :mod:`flask.ext.admin.contrib.sqla.filters` for usage example.
     """
     def _inner(func):
         func._converter_for = args
